@@ -185,7 +185,7 @@ OvsDoEncapStt(POVS_VPORT_ENTRY vport,
         if ((innerFrameLen > OVS_MAX_STT_PACKET_LENGTH) ||
             (layers->l4Offset > OVS_MAX_STT_L4_OFFSET_LENGTH)) {
             *newNbl = OvsTcpSegmentNBL(switchContext, curNbl, layers,
-                mss - headRoom, headRoom);
+                                       mss - headRoom, headRoom, FALSE);
             if (*newNbl == NULL) {
                 OVS_LOG_ERROR("Unable to segment NBL");
                 return NDIS_STATUS_FAILURE;
@@ -695,6 +695,7 @@ OvsSttReassemble(POVS_SWITCH_CONTEXT switchContext,
         entry->timeout = currentTime + STT_ENTRY_TIMEOUT;
 
         if (segOffset == 0) {
+            ASSERT(sttHdr);
             entry->sttHdr = *sttHdr;
         }
 
@@ -722,6 +723,7 @@ OvsSttReassemble(POVS_SWITCH_CONTEXT switchContext,
         }
 
         if (segOffset == 0) {
+            ASSERT(sttHdr);
             pktFragEntry->sttHdr = *sttHdr;
         }
         if (ipHdr->ecn == IP_ECN_CE) {
@@ -855,8 +857,8 @@ OvsDecapSetOffloads(PNET_BUFFER_LIST *curNbl,
             tcpHdr = (TCPHdr *)(buf + layers->l4Offset);
 
             tcpHdr->check = IPPseudoChecksum(&ipHdr->saddr,
-                                                (uint32 *)&ipHdr->daddr,
-                                                IPPROTO_TCP, 0);
+                                             (uint32 *)&ipHdr->daddr,
+                                             IPPROTO_TCP, 0);
         } else {
             IPv6Hdr *ipHdr;
             TCPHdr *tcpHdr;
@@ -1019,7 +1021,7 @@ OvsDecapStt(POVS_SWITCH_CONTEXT switchContext,
                 innerIpHdr->check = IPChecksum((UINT8 *)innerIpHdr,
                                                 innerIpHdr->ihl * 4, 0);
             } else {
-                status = NDIS_STATUS_RESOURCES;
+                status = NDIS_STATUS_INVALID_PACKET;
                 goto dropNbl;
             }
         } else if (layers.isIPv6) {
