@@ -27,6 +27,7 @@
 #include "User.h"
 #include "Vport.h"
 #include "Vxlan.h"
+#include "Geneve.h"
 
 #ifdef OVS_DBG_MOD
 #undef OVS_DBG_MOD
@@ -81,8 +82,6 @@ static NTSTATUS CreateNetlinkMesgForNetdev(POVS_VPORT_EXT_INFO info,
                                            PVOID outBuffer,
                                            UINT32 outBufLen,
                                            int dpIfIndex);
-static POVS_VPORT_ENTRY OvsFindVportByHvNameW(POVS_SWITCH_CONTEXT switchContext,
-                                              PWSTR wsName, SIZE_T wstrSize);
 static VOID UpdateSwitchCtxWithVport(POVS_SWITCH_CONTEXT switchContext,
                                      POVS_VPORT_ENTRY vport, BOOLEAN newPort);
 static NTSTATUS OvsRemoveTunnelVport(POVS_USER_PARAMS_CONTEXT usrParamsCtx,
@@ -96,8 +95,11 @@ static VOID OvsTunnelVportPendingInit(PVOID context,
 static VOID OvsTunnelVportPendingRemove(PVOID context,
                                         NTSTATUS status,
                                         UINT32 *replyLen);
-static NTSTATUS GetNICAlias(GUID *netCfgInstanceId,
+static NTSTATUS GetNICAlias(PNDIS_SWITCH_NIC_PARAMETERS nicParam,
                             IF_COUNTED_STRING *portFriendlyName);
+static NTSTATUS OvsConvertIfCountedStrToAnsiStr(PIF_COUNTED_STRING wStr,
+                                                CHAR *str,
+                                                UINT16 maxStrLen);
 
 /*
  * --------------------------------------------------------------------------
@@ -291,7 +293,15 @@ HvDeletePort(POVS_SWITCH_CONTEXT switchContext,
      * delete will delete the vport.
     */
     if (vport) {
+        OVS_VPORT_EVENT_ENTRY event;
+
+        event.portNo = vport->portNo;
+        event.ovsType = vport->ovsType;
+        event.upcallPid = vport->upcallPid;
+        RtlCopyMemory(&event.ovsName, &vport->ovsName, sizeof event.ovsName);
+        event.type = OVS_EVENT_LINK_DOWN;
         OvsRemoveAndDeleteVport(NULL, switchContext, vport, TRUE, FALSE);
+        OvsPostVportEvent(&event);
     } else {
         OVS_LOG_WARN("Vport not present.");
     }
@@ -331,7 +341,7 @@ HvCreateNic(POVS_SWITCH_CONTEXT switchContext,
 
     if (OvsIsInternalNIC(nicParam->NicType) ||
         OvsIsRealExternalNIC(nicParam->NicType, nicParam->NicIndex)) {
-        GetNICAlias(&nicParam->NetCfgInstanceId, &portFriendlyName);
+        GetNICAlias(nicParam, &portFriendlyName);
     }
 
     NdisAcquireRWLockWrite(switchContext->dispatchLock, &lockState, 0);
@@ -341,26 +351,46 @@ HvCreateNic(POVS_SWITCH_CONTEXT switchContext,
      * from the parent external port.
      */
     if (OvsIsRealExternalNIC(nicParam->NicType, nicParam->NicIndex)) {
-        NDIS_SWITCH_PORT_PARAMETERS portParam;
-        POVS_VPORT_ENTRY virtExtVport =
-            (POVS_VPORT_ENTRY)switchContext->virtualExternalVport;
-
-        ASSERT(virtExtVport);
+        /* The VPORT can be bound to OVS datapath already. Search for it
+         * using its friendly name and if not found allocate a new port
+         */
         ASSERT(OvsFindVportByPortIdAndNicIndex(switchContext,
                                                nicParam->PortId,
                                                nicParam->NicIndex) == NULL);
-        OvsCopyPortParamsFromVport(virtExtVport, &portParam);
+        char convertString[256];
+        RtlZeroMemory(convertString, 256);
         NdisReleaseRWLock(switchContext->dispatchLock, &lockState);
-        status = HvCreatePort(switchContext, &portParam,
-                              nicParam->NicIndex);
+        status = OvsConvertIfCountedStrToAnsiStr(&portFriendlyName,
+                                                 convertString,
+                                                 OVS_MAX_PORT_NAME_LENGTH);
         NdisAcquireRWLockWrite(switchContext->dispatchLock, &lockState, 0);
         if (status != NDIS_STATUS_SUCCESS) {
             goto add_nic_done;
         }
+        POVS_VPORT_ENTRY ovsVport = OvsFindVportByOvsName(switchContext,
+                                                          convertString);
+        if (ovsVport != NULL) {
+            UpdateSwitchCtxWithVport(switchContext, ovsVport, FALSE);
+        } else {
+            NDIS_SWITCH_PORT_PARAMETERS portParam;
+            POVS_VPORT_ENTRY virtExtVport =
+                (POVS_VPORT_ENTRY)switchContext->virtualExternalVport;
+
+            ASSERT(virtExtVport);
+            OvsCopyPortParamsFromVport(virtExtVport, &portParam);
+            NdisReleaseRWLock(switchContext->dispatchLock, &lockState);
+            status = HvCreatePort(switchContext, &portParam,
+                                  nicParam->NicIndex);
+            NdisAcquireRWLockWrite(switchContext->dispatchLock, &lockState, 0);
+            if (status != NDIS_STATUS_SUCCESS) {
+                goto add_nic_done;
+            }
+        }
     }
 
     vport = OvsFindVportByPortIdAndNicIndex(switchContext, nicParam->PortId,
                                             nicParam->NicIndex);
+
     if (vport == NULL) {
         OVS_LOG_ERROR("Create NIC without Switch Port,"
                       " PortId: %x, NicIndex: %d",
@@ -396,7 +426,6 @@ HvConnectNic(POVS_SWITCH_CONTEXT switchContext,
 {
     LOCK_STATE_EX lockState;
     POVS_VPORT_ENTRY vport;
-    UINT32 portNo;
 
     VPORT_NIC_ENTER(nicParam);
 
@@ -422,12 +451,11 @@ HvConnectNic(POVS_SWITCH_CONTEXT switchContext,
 
     vport->ovsState = OVS_STATE_CONNECTED;
     vport->nicState = NdisSwitchNicStateConnected;
-    portNo = vport->portNo;
 
     NdisReleaseRWLock(switchContext->dispatchLock, &lockState);
 
     if (nicParam->NicType == NdisSwitchNicTypeInternal) {
-        OvsInternalAdapterUp(&nicParam->NetCfgInstanceId);
+        OvsInternalAdapterUp(vport->portNo, &vport->netCfgInstanceId);
     }
 
 done:
@@ -464,7 +492,7 @@ HvUpdateNic(POVS_SWITCH_CONTEXT switchContext,
     /* GetNICAlias() must be called outside of a lock. */
     if (nicParam->NicType == NdisSwitchNicTypeInternal ||
         OvsIsRealExternalNIC(nicParam->NicType, nicParam->NicIndex)) {
-        GetNICAlias(&nicParam->NetCfgInstanceId, &portFriendlyName);
+        GetNICAlias(nicParam, &portFriendlyName);
         aliasLookup = TRUE;
     }
 
@@ -524,14 +552,14 @@ HvUpdateNic(POVS_SWITCH_CONTEXT switchContext,
     vport->numaNodeId = nicParam->NumaNodeId;
 
     if (nameChanged) {
-        OVS_EVENT_ENTRY event;
-        event.portNo = vport->portNo;
-        event.ovsType = vport->ovsType;
-        event.upcallPid = vport->upcallPid;
-        RtlCopyMemory(&event.ovsName, &vport->ovsName, sizeof event.ovsName);
-        event.type = OVS_EVENT_LINK_DOWN;
+        OVS_VPORT_EVENT_ENTRY evt;
+        evt.portNo = vport->portNo;
+        evt.ovsType = vport->ovsType;
+        evt.upcallPid = vport->upcallPid;
+        RtlCopyMemory(&evt.ovsName, &vport->ovsName, sizeof evt.ovsName);
+        evt.type = OVS_EVENT_LINK_DOWN;
         OvsRemoveAndDeleteVport(NULL, switchContext, vport, FALSE, TRUE);
-        OvsPostEvent(&event);
+        OvsPostVportEvent(&evt);
     }
 
     NdisReleaseRWLock(switchContext->dispatchLock, &lockState);
@@ -560,7 +588,7 @@ HvDisconnectNic(POVS_SWITCH_CONTEXT switchContext,
     POVS_VPORT_ENTRY vport;
     LOCK_STATE_EX lockState;
     BOOLEAN isInternalPort = FALSE;
-    OVS_EVENT_ENTRY event;
+    OVS_VPORT_EVENT_ENTRY event;
 
     VPORT_NIC_ENTER(nicParam);
 
@@ -602,12 +630,14 @@ HvDisconnectNic(POVS_SWITCH_CONTEXT switchContext,
      */
     if (OvsIsRealExternalVport(vport)) {
         OvsRemoveAndDeleteVport(NULL, switchContext, vport, FALSE, TRUE);
-        OvsPostEvent(&event);
+        OvsPostVportEvent(&event);
     }
     NdisReleaseRWLock(switchContext->dispatchLock, &lockState);
 
     if (isInternalPort) {
-        OvsInternalAdapterDown();
+        OvsInternalAdapterDown(vport->portNo, vport->netCfgInstanceId);
+        OvsRemoveAndDeleteVport(NULL, switchContext, vport, TRUE, TRUE);
+        OvsPostVportEvent(&event);
     }
 
 done:
@@ -683,9 +713,9 @@ OvsFindVportByPortNo(POVS_SWITCH_CONTEXT switchContext,
 
 
 POVS_VPORT_ENTRY
-OvsFindTunnelVportByDstPort(POVS_SWITCH_CONTEXT switchContext,
-                            UINT16 dstPort,
-                            OVS_VPORT_TYPE ovsPortType)
+OvsFindTunnelVportByDstPortAndType(POVS_SWITCH_CONTEXT switchContext,
+                                   UINT16 dstPort,
+                                   OVS_VPORT_TYPE ovsPortType)
 {
     POVS_VPORT_ENTRY vport;
     PLIST_ENTRY head, link;
@@ -702,6 +732,42 @@ OvsFindTunnelVportByDstPort(POVS_SWITCH_CONTEXT switchContext,
     return NULL;
 }
 
+POVS_VPORT_ENTRY
+OvsFindTunnelVportByDstPortAndNWProto(POVS_SWITCH_CONTEXT switchContext,
+                                      UINT16 dstPort,
+                                      UINT8 nwProto)
+{
+    POVS_VPORT_ENTRY vport;
+    PLIST_ENTRY head, link;
+    UINT32 hash = OvsJhashBytes((const VOID *)&dstPort, sizeof(dstPort),
+                                OVS_HASH_BASIS);
+    head = &(switchContext->tunnelVportsArray[hash & OVS_VPORT_MASK]);
+    LIST_FORALL(head, link) {
+        vport = CONTAINING_RECORD(link, OVS_VPORT_ENTRY, tunnelVportLink);
+        if (GetPortFromPriv(vport) == dstPort) {
+            switch (nwProto) {
+            case IPPROTO_UDP:
+                if (vport->ovsType != OVS_VPORT_TYPE_GENEVE &&
+                    vport->ovsType != OVS_VPORT_TYPE_VXLAN) {
+                    continue;
+                }
+                break;
+            case IPPROTO_TCP:
+                if (vport->ovsType != OVS_VPORT_TYPE_STT) {
+                    continue;
+                }
+                break;
+            case IPPROTO_GRE:
+                break;
+            default:
+                continue;
+            }
+            return vport;
+        }
+    }
+    return NULL;
+}
+
 POVS_VPORT_ENTRY
 OvsFindTunnelVportByPortType(POVS_SWITCH_CONTEXT switchContext,
                              OVS_VPORT_TYPE ovsPortType)
@@ -827,10 +893,6 @@ OvsFindVportByPortIdAndNicIndex(POVS_SWITCH_CONTEXT switchContext,
             portId == switchContext->virtualExternalPortId &&
             index == switchContext->virtualExternalVport->nicIndex) {
         return (POVS_VPORT_ENTRY)switchContext->virtualExternalVport;
-    } else if (switchContext->internalVport &&
-               portId == switchContext->internalPortId &&
-               index == switchContext->internalVport->nicIndex) {
-        return (POVS_VPORT_ENTRY)switchContext->internalVport;
     } else {
         PLIST_ENTRY head, link;
         POVS_VPORT_ENTRY vport;
@@ -847,6 +909,12 @@ OvsFindVportByPortIdAndNicIndex(POVS_SWITCH_CONTEXT switchContext,
     }
 }
 
+BOOLEAN OvsIsExternalVportByPortId(POVS_SWITCH_CONTEXT switchContext,
+                                   NDIS_SWITCH_PORT_ID portId)
+{
+    return (portId == switchContext->virtualExternalPortId);
+}
+
 POVS_VPORT_ENTRY
 OvsAllocateVport(VOID)
 {
@@ -877,7 +945,6 @@ OvsInitVportWithPortParam(POVS_VPORT_ENTRY vport,
     vport->portId = portParam->PortId;
     vport->nicState = NdisSwitchNicStateUnknown;
     vport->isExternal = FALSE;
-    vport->isBridgeInternal = FALSE;
 
     switch (vport->portType) {
     case NdisSwitchPortTypeExternal:
@@ -919,24 +986,25 @@ OvsInitVportWithNicParam(POVS_SWITCH_CONTEXT switchContext,
                          PNDIS_SWITCH_NIC_PARAMETERS nicParam)
 {
     ASSERT(vport->portId == nicParam->PortId);
-    ASSERT(vport->ovsState == OVS_STATE_PORT_CREATED);
 
     UNREFERENCED_PARAMETER(switchContext);
 
     RtlCopyMemory(vport->permMacAddress, nicParam->PermanentMacAddress,
-                  sizeof (nicParam->PermanentMacAddress));
+                  sizeof (vport->permMacAddress));
     RtlCopyMemory(vport->currMacAddress, nicParam->CurrentMacAddress,
-                  sizeof (nicParam->CurrentMacAddress));
+                  sizeof (vport->currMacAddress));
 
     if (nicParam->NicType == NdisSwitchNicTypeSynthetic ||
         nicParam->NicType == NdisSwitchNicTypeEmulated) {
         RtlCopyMemory(vport->vmMacAddress, nicParam->VMMacAddress,
-                      sizeof (nicParam->VMMacAddress));
+                      sizeof (vport->vmMacAddress));
         RtlCopyMemory(&vport->vmName, &nicParam->VmName,
                       sizeof (nicParam->VmName));
     } else {
         RtlCopyMemory(&vport->netCfgInstanceId, &nicParam->NetCfgInstanceId,
                       sizeof (nicParam->NetCfgInstanceId));
+        RtlCopyMemory(&vport->nicFriendlyName, &nicParam->NicFriendlyName,
+                      sizeof (nicParam->NicFriendlyName));
     }
     RtlCopyMemory(&vport->nicName, &nicParam->NicName,
                   sizeof (nicParam->NicName));
@@ -998,7 +1066,6 @@ OvsInitTunnelVport(PVOID userContext,
     POVS_USER_PARAMS_CONTEXT usrParamsCtx =
         (POVS_USER_PARAMS_CONTEXT)userContext;
 
-    vport->isBridgeInternal = FALSE;
     vport->ovsType = ovsType;
     vport->ovsState = OVS_STATE_PORT_CREATED;
     switch (ovsType) {
@@ -1034,45 +1101,38 @@ OvsInitTunnelVport(PVOID userContext,
     case OVS_VPORT_TYPE_STT:
         status = OvsInitSttTunnel(vport, dstPort);
         break;
+    case OVS_VPORT_TYPE_GENEVE:
+        status = OvsInitGeneveTunnel(vport, dstPort);
+        break;
     default:
         ASSERT(0);
     }
     return status;
 }
 
-/*
- * --------------------------------------------------------------------------
- * Initializes a bridge internal vport ie. a port of type
- * OVS_VPORT_TYPE_INTERNAL but not present on the Hyper-V switch.
- * --------------------------------------------------------------------------
- */
-NTSTATUS
-OvsInitBridgeInternalVport(POVS_VPORT_ENTRY vport)
-{
-    vport->isBridgeInternal = TRUE;
-    vport->ovsType = OVS_VPORT_TYPE_INTERNAL;
-    /* Mark the status to be connected, since there is no other initialization
-     * for this port. */
-    vport->ovsState = OVS_STATE_CONNECTED;
-    return STATUS_SUCCESS;
-}
-
 /*
  * --------------------------------------------------------------------------
  * For external and internal vports 'portFriendlyName' parameter, provided by
- * Hyper-V, is overwritten with the interface alias name.
+  * Hyper-V, is overwritten with the interface alias name and NIC friendly name
+  * equivalent.
  * --------------------------------------------------------------------------
  */
 static NTSTATUS
-GetNICAlias(GUID *netCfgInstanceId,
+GetNICAlias(PNDIS_SWITCH_NIC_PARAMETERS nicParam,
             IF_COUNTED_STRING *portFriendlyName)
 {
     NTSTATUS status = STATUS_SUCCESS;
-    WCHAR interfaceName[IF_MAX_STRING_SIZE] = { 0 };
-    NET_LUID interfaceLuid = { 0 };
-    size_t len = 0;
+    WCHAR interfaceName[IF_MAX_STRING_SIZE + 1];
+    NET_LUID interfaceLuid;
+    size_t len;
 
-    status = ConvertInterfaceGuidToLuid(netCfgInstanceId,
+    if (nicParam->NicType == NdisSwitchNicTypeInternal) {
+        RtlCopyMemory(portFriendlyName, &nicParam->NicFriendlyName,
+                      sizeof nicParam->NicFriendlyName);
+    return status;
+    }
+
+    status = ConvertInterfaceGuidToLuid(&nicParam->NetCfgInstanceId,
                                         &interfaceLuid);
     if (status == STATUS_SUCCESS) {
         /*
@@ -1121,14 +1181,12 @@ UpdateSwitchCtxWithVport(POVS_SWITCH_CONTEXT switchContext,
         if (vport->nicIndex == 0) {
             switchContext->virtualExternalPortId = vport->portId;
             switchContext->virtualExternalVport = vport;
-        } else {
+        } else if (newPort == TRUE) {
             switchContext->numPhysicalNics++;
         }
         break;
     case NdisSwitchPortTypeInternal:
-        ASSERT(vport->isBridgeInternal == FALSE);
-        switchContext->internalPortId = vport->portId;
-        switchContext->internalVport = vport;
+        switchContext->countInternalVports++;
         break;
     case NdisSwitchPortTypeSynthetic:
     case NdisSwitchPortTypeEmulated:
@@ -1177,6 +1235,7 @@ InitOvsVportCommon(POVS_SWITCH_CONTEXT switchContext,
     case OVS_VPORT_TYPE_GRE:
     case OVS_VPORT_TYPE_VXLAN:
     case OVS_VPORT_TYPE_STT:
+    case OVS_VPORT_TYPE_GENEVE:
     {
         UINT16 dstPort = GetPortFromPriv(vport);
         hash = OvsJhashBytes(&dstPort,
@@ -1188,10 +1247,6 @@ InitOvsVportCommon(POVS_SWITCH_CONTEXT switchContext,
         switchContext->numNonHvVports++;
         break;
     }
-    case OVS_VPORT_TYPE_INTERNAL:
-        if (vport->isBridgeInternal) {
-            switchContext->numNonHvVports++;
-        }
     default:
         break;
     }
@@ -1242,14 +1297,12 @@ OvsRemoveAndDeleteVport(PVOID usrParamsContext,
 
     switch (vport->ovsType) {
     case OVS_VPORT_TYPE_INTERNAL:
-        if (!vport->isBridgeInternal) {
-            if (hvDelete && vport->isAbsentOnHv == FALSE) {
-                switchContext->internalPortId = 0;
-                switchContext->internalVport = NULL;
-                OvsInternalAdapterDown();
-            }
-            hvSwitchPort = TRUE;
+        if (hvDelete && vport->isAbsentOnHv == FALSE) {
+            switchContext->countInternalVports--;
+            ASSERT(switchContext->countInternalVports >= 0);
+            OvsInternalAdapterDown(vport->portNo, vport->netCfgInstanceId);
         }
+        hvSwitchPort = TRUE;
         break;
     case OVS_VPORT_TYPE_VXLAN:
     {
@@ -1260,6 +1313,9 @@ OvsRemoveAndDeleteVport(PVOID usrParamsContext,
             return status;
         }
     }
+    case OVS_VPORT_TYPE_GENEVE:
+        OvsCleanupGeneveTunnel(vport);
+        break;
     case OVS_VPORT_TYPE_STT:
         OvsCleanupSttTunnel(vport);
         break;
@@ -1321,9 +1377,7 @@ OvsRemoveAndDeleteVport(PVOID usrParamsContext,
         InitializeListHead(&vport->ovsNameLink);
         RemoveEntryList(&vport->portNoLink);
         InitializeListHead(&vport->portNoLink);
-        if (OVS_VPORT_TYPE_VXLAN == vport->ovsType ||
-            OVS_VPORT_TYPE_STT == vport->ovsType   ||
-            OVS_VPORT_TYPE_GRE == vport->ovsType) {
+        if (OvsIsTunnelVportType(vport->ovsType)) {
             RemoveEntryList(&vport->tunnelVportLink);
             InitializeListHead(&vport->tunnelVportLink);
         }
@@ -1509,14 +1563,13 @@ OvsClearAllSwitchVports(POVS_SWITCH_CONTEXT switchContext)
             POVS_VPORT_ENTRY vport;
             vport = CONTAINING_RECORD(link, OVS_VPORT_ENTRY, portNoLink);
             ASSERT(OvsIsTunnelVportType(vport->ovsType) ||
-                   (vport->ovsType == OVS_VPORT_TYPE_INTERNAL &&
-                    vport->isBridgeInternal) || vport->isAbsentOnHv == TRUE);
+                   vport->isAbsentOnHv == TRUE);
             OvsRemoveAndDeleteVport(NULL, switchContext, vport, TRUE, TRUE);
         }
     }
 
     ASSERT(switchContext->virtualExternalVport == NULL);
-    ASSERT(switchContext->internalVport == NULL);
+    ASSERT(switchContext->countInternalVports == 0);
 }
 
 
@@ -1565,7 +1618,6 @@ OvsGetExtInfoIoctl(POVS_VPORT_GET vportGet,
                    POVS_VPORT_EXT_INFO extInfo)
 {
     POVS_VPORT_ENTRY vport;
-    size_t len;
     LOCK_STATE_EX lockState;
     NTSTATUS status = STATUS_SUCCESS;
     BOOLEAN doConvert = FALSE;
@@ -1573,7 +1625,6 @@ OvsGetExtInfoIoctl(POVS_VPORT_GET vportGet,
     RtlZeroMemory(extInfo, sizeof (POVS_VPORT_EXT_INFO));
     NdisAcquireRWLockRead(gOvsSwitchContext->dispatchLock, &lockState, 0);
     if (vportGet->portNo == 0) {
-        StringCbLengthA(vportGet->name, OVS_MAX_PORT_NAME_LENGTH - 1, &len);
         vport = OvsFindVportByHvNameA(gOvsSwitchContext, vportGet->name);
         if (vport == NULL) {
             /* If the port is not a Hyper-V port and it has been added earlier,
@@ -1633,7 +1684,7 @@ OvsGetExtInfoIoctl(POVS_VPORT_GET vportGet,
                                                  OVS_MAX_PORT_NAME_LENGTH);
         if (status != STATUS_SUCCESS) {
             OVS_LOG_INFO("Fail to convert NIC name.");
-            extInfo->vmUUID[0] = 0;
+            extInfo->name[0] = 0;
         }
 
         status = OvsConvertIfCountedStrToAnsiStr(&vport->vmName,
@@ -1722,8 +1773,9 @@ cleanup:
         POVS_MESSAGE_ERROR msgError = (POVS_MESSAGE_ERROR)
             usrParamsCtx->outputBuffer;
 
-        NlBuildErrorMsg(msgIn, msgError, nlError);
-        *replyLen = msgError->nlMsg.nlmsgLen;
+        ASSERT(msgError);
+        NlBuildErrorMsg(msgIn, msgError, nlError, replyLen);
+        ASSERT(*replyLen != 0);
     }
 
     return STATUS_SUCCESS;
@@ -2081,8 +2133,9 @@ Cleanup:
         POVS_MESSAGE_ERROR msgError = (POVS_MESSAGE_ERROR)
             usrParamsCtx->outputBuffer;
 
-        NlBuildErrorMsg(msgIn, msgError, nlError);
-        *replyLen = msgError->nlMsg.nlmsgLen;
+        ASSERT(msgError);
+        NlBuildErrorMsg(msgIn, msgError, nlError, replyLen);
+        ASSERT(*replyLen != 0);
     }
 
     return STATUS_SUCCESS;
@@ -2196,12 +2249,12 @@ OvsNewVportCmdHandler(POVS_USER_PARAMS_CONTEXT usrParamsCtx,
         goto Cleanup;
     }
 
-    if (portType == OVS_VPORT_TYPE_NETDEV) {
-        /* External ports can also be looked up like VIF ports. */
+    if (portType == OVS_VPORT_TYPE_NETDEV ||
+        portType == OVS_VPORT_TYPE_INTERNAL) {
+        /* External and internal ports can also be looked up like VIF ports. */
         vport = OvsFindVportByHvNameA(gOvsSwitchContext, portName);
     } else {
-        ASSERT(OvsIsTunnelVportType(portType) ||
-               portType == OVS_VPORT_TYPE_INTERNAL);
+        ASSERT(OvsIsTunnelVportType(portType));
 
         vport = (POVS_VPORT_ENTRY)OvsAllocateVport();
         if (vport == NULL) {
@@ -2212,15 +2265,23 @@ OvsNewVportCmdHandler(POVS_USER_PARAMS_CONTEXT usrParamsCtx,
 
         if (OvsIsTunnelVportType(portType)) {
             UINT16 transportPortDest = 0;
+            UINT8 nwProto = IPPROTO_NONE;
+            POVS_VPORT_ENTRY dupVport;
 
             switch (portType) {
             case OVS_VPORT_TYPE_GRE:
+                nwProto = IPPROTO_GRE;
                 break;
             case OVS_VPORT_TYPE_VXLAN:
                 transportPortDest = VXLAN_UDP_PORT;
+                nwProto = IPPROTO_UDP;
+                break;
+            case OVS_VPORT_TYPE_GENEVE:
+                transportPortDest = GENEVE_UDP_PORT;
                 break;
             case OVS_VPORT_TYPE_STT:
                 transportPortDest = STT_TCP_PORT;
+                nwProto = IPPROTO_TCP;
                 break;
             default:
                 nlError = NL_ERROR_INVAL;
@@ -2235,14 +2296,28 @@ OvsNewVportCmdHandler(POVS_USER_PARAMS_CONTEXT usrParamsCtx,
                 }
             }
 
+            /*
+             * We don't allow two tunnel ports on identical N/W protocol and
+             * L4 port number. This is applicable even if the two ports are of
+             * different tunneling types.
+             */
+            dupVport =
+                OvsFindTunnelVportByDstPortAndNWProto(gOvsSwitchContext,
+                                                      transportPortDest,
+                                                      nwProto);
+            if (dupVport) {
+                OVS_LOG_ERROR("Vport for N/W proto and port already exists,"
+                    " type: %u, dst port: %u, name: %s", dupVport->ovsType,
+                    transportPortDest, dupVport->ovsName);
+                goto Cleanup;
+            }
+
             status = OvsInitTunnelVport(usrParamsCtx,
                                         vport,
                                         portType,
                                         transportPortDest);
 
             nlError = NlMapStatusToNlErr(status);
-        } else {
-            OvsInitBridgeInternalVport(vport);
         }
 
         vportInitialized = TRUE;
@@ -2309,6 +2384,8 @@ OvsNewVportCmdHandler(POVS_USER_PARAMS_CONTEXT usrParamsCtx,
                                    gOvsSwitchContext->dpNo);
 
     *replyLen = msgOut->nlMsg.nlmsgLen;
+    OVS_LOG_INFO("Created new vport, name: %s, type: %u", vport->ovsName,
+                 vport->ovsType);
 
 Cleanup:
     NdisReleaseRWLock(gOvsSwitchContext->dispatchLock, &lockState);
@@ -2327,6 +2404,9 @@ Cleanup:
                     case OVS_VPORT_TYPE_STT:
                         OvsCleanupSttTunnel(vport);
                         break;
+                    case OVS_VPORT_TYPE_GENEVE:
+                        OvsCleanupGeneveTunnel(vport);
+                        break;
                     default:
                         ASSERT(!"Invalid tunnel port type");
                     }
@@ -2335,8 +2415,9 @@ Cleanup:
             OvsFreeMemoryWithTag(vport, OVS_VPORT_POOL_TAG);
         }
 
-        NlBuildErrorMsg(msgIn, msgError, nlError);
-        *replyLen = msgError->nlMsg.nlmsgLen;
+        ASSERT(msgError);
+        NlBuildErrorMsg(msgIn, msgError, nlError, replyLen);
+        ASSERT(*replyLen != 0);
     }
 
     return (status == STATUS_PENDING) ? STATUS_PENDING : STATUS_SUCCESS;
@@ -2445,8 +2526,9 @@ Cleanup:
         POVS_MESSAGE_ERROR msgError = (POVS_MESSAGE_ERROR)
             usrParamsCtx->outputBuffer;
 
-        NlBuildErrorMsg(msgIn, msgError, nlError);
-        *replyLen = msgError->nlMsg.nlmsgLen;
+        ASSERT(msgError);
+        NlBuildErrorMsg(msgIn, msgError, nlError, replyLen);
+        ASSERT(*replyLen != 0);
     }
 
     return STATUS_SUCCESS;
@@ -2537,8 +2619,9 @@ Cleanup:
         POVS_MESSAGE_ERROR msgError = (POVS_MESSAGE_ERROR)
             usrParamsCtx->outputBuffer;
 
-        NlBuildErrorMsg(msgIn, msgError, nlError);
-        *replyLen = msgError->nlMsg.nlmsgLen;
+        ASSERT(msgError);
+        NlBuildErrorMsg(msgIn, msgError, nlError, replyLen);
+        ASSERT(*replyLen != 0);
     }
 
     return (status == STATUS_PENDING) ? STATUS_PENDING : STATUS_SUCCESS;
@@ -2572,9 +2655,9 @@ OvsTunnelVportPendingRemove(PVOID context,
             *replyLen = msgOut->nlMsg.nlmsgLen;
         } else {
             POVS_MESSAGE_ERROR msgError = (POVS_MESSAGE_ERROR)msgOut;
-
-            NlBuildErrorMsg(msgIn, msgError, nlError);
-            *replyLen = msgError->nlMsg.nlmsgLen;
+            ASSERT(msgError);
+            NlBuildErrorMsg(msgIn, msgError, nlError, replyLen);
+            ASSERT(*replyLen != 0);
         }
     }
 
@@ -2713,12 +2796,13 @@ OvsTunnelVportPendingInit(PVOID context,
     } while (error);
 
     if (error) {
-        POVS_MESSAGE_ERROR msgError = (POVS_MESSAGE_ERROR) msgOut;
+        POVS_MESSAGE_ERROR msgError = (POVS_MESSAGE_ERROR)msgOut;
 
         OvsCleanupVxlanTunnel(NULL, vport, NULL, NULL);
         OvsFreeMemory(vport);
 
-        NlBuildErrorMsg(msgIn, msgError, nlError);
-        *replyLen = msgError->nlMsg.nlmsgLen;
+        ASSERT(msgError);
+        NlBuildErrorMsg(msgIn, msgError, nlError, replyLen);
+        ASSERT(*replyLen != 0);
     }
 }
