@@ -457,7 +457,6 @@ static int lisp_open(struct net_device *dev)
 
 	rcu_assign_pointer(lisp->sock, sock);
 	/* Mark socket as an encapsulation socket */
-	memset(&tunnel_cfg, 0, sizeof(tunnel_cfg));
 	tunnel_cfg.sk_user_data = dev;
 	tunnel_cfg.encap_type = 1;
 	tunnel_cfg.encap_rcv = lisp_rcv;
@@ -581,7 +580,11 @@ static void lisp_setup(struct net_device *dev)
 
 	dev->netdev_ops = &lisp_netdev_ops;
 	dev->ethtool_ops = &lisp_ethtool_ops;
+#ifndef HAVE_NEEDS_FREE_NETDEV
 	dev->destructor = free_netdev;
+#else
+	dev->needs_free_netdev = true;
+#endif
 
 	SET_NETDEV_DEVTYPE(dev, &lisp_type);
 
