@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016 Nicira, Inc.
+ * Copyright (c) 2009-2017 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -269,9 +269,9 @@ cls_rule_equal(const struct cls_rule *a, const struct cls_rule *b)
 /* Appends a string describing 'rule' to 's'. */
 void
 cls_rule_format(const struct cls_rule *rule, const struct tun_table *tun_table,
-                struct ds *s)
+                const struct ofputil_port_map *port_map, struct ds *s)
 {
-    minimatch_format(&rule->match, tun_table, s, rule->priority);
+    minimatch_format(&rule->match, tun_table, port_map, s, rule->priority);
 }
 
 /* Returns true if 'rule' matches every packet, false otherwise. */
@@ -793,8 +793,6 @@ check_priority:
             && --subtable->max_count == 0) {
             /* Find the new 'max_priority' and 'max_count'. */
             int max_priority = INT_MIN;
-            struct cls_match *head;
-
             CMAP_FOR_EACH (head, cmap_node, &subtable->rules) {
                 if (head->priority > max_priority) {
                     max_priority = head->priority;
