@@ -31,7 +31,10 @@ dp_packet_init__(struct dp_packet *b, size_t allocated, enum dp_packet_source so
     dp_packet_reset_offsets(b);
     pkt_metadata_init(&b->md, 0);
     dp_packet_rss_invalidate(b);
+    dp_packet_mbuf_init(b);
     dp_packet_reset_cutlen(b);
+    /* By default assume the packet type to be Ethernet. */
+    b->packet_type = htonl(PT_ETH);
 }
 
 static void
@@ -171,6 +174,7 @@ dp_packet_clone_with_headroom(const struct dp_packet *buffer, size_t headroom)
     new_buffer->l4_ofs = buffer->l4_ofs;
     new_buffer->md = buffer->md;
     new_buffer->cutlen = buffer->cutlen;
+    new_buffer->packet_type = buffer->packet_type;
 #ifdef DPDK_NETDEV
     new_buffer->mbuf.ol_flags = buffer->mbuf.ol_flags;
 #else
