@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014 Nicira, Inc.
+ * Copyright (c) 2014, 2017 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -29,8 +29,6 @@
 
 VLOG_DEFINE_THIS_MODULE(ovs_rcu);
 
-#define MIN_CBS 16
-
 struct ovsrcu_cb {
     void (*function)(void *aux);
     void *aux;
@@ -38,8 +36,7 @@ struct ovsrcu_cb {
 
 struct ovsrcu_cbset {
     struct ovs_list list_node;
-    struct ovsrcu_cb *cbs;
-    size_t n_allocated;
+    struct ovsrcu_cb cbs[16];
     int n_cbs;
 };
 
@@ -210,8 +207,7 @@ ovsrcu_synchronize(void)
         ovs_mutex_lock(&ovsrcu_threads_mutex);
         LIST_FOR_EACH (perthread, list_node, &ovsrcu_threads) {
             if (perthread->seqno <= target_seqno) {
-                ovs_strlcpy(stalled_thread, perthread->name,
-                            sizeof stalled_thread);
+                ovs_strlcpy_arrays(stalled_thread, perthread->name);
                 done = false;
                 break;
             }
@@ -265,19 +261,16 @@ ovsrcu_postpone__(void (*function)(void *aux), void *aux)
     cbset = perthread->cbset;
     if (!cbset) {
         cbset = perthread->cbset = xmalloc(sizeof *perthread->cbset);
-        cbset->cbs = xmalloc(MIN_CBS * sizeof *cbset->cbs);
-        cbset->n_allocated = MIN_CBS;
         cbset->n_cbs = 0;
     }
 
-    if (cbset->n_cbs == cbset->n_allocated) {
-        cbset->cbs = x2nrealloc(cbset->cbs, &cbset->n_allocated,
-                                sizeof *cbset->cbs);
-    }
-
     cb = &cbset->cbs[cbset->n_cbs++];
     cb->function = function;
     cb->aux = aux;
+
+    if (cbset->n_cbs >= ARRAY_SIZE(cbset->cbs)) {
+        ovsrcu_flush_cbset(perthread);
+    }
 }
 
 static bool
@@ -299,7 +292,6 @@ ovsrcu_call_postponed(void)
         for (cb = cbset->cbs; cb < &cbset->cbs[cbset->n_cbs]; cb++) {
             cb->function(cb->aux);
         }
-        free(cbset->cbs);
         free(cbset);
     }
 
