@@ -128,7 +128,7 @@ send_bogus_packet_ins(struct fail_open *fo)
         .oam = OAM_PACKET_IN,
         .pin = {
             .up = {
-                .public = {
+                .base = {
                     .packet = dp_packet_data(&b),
                     .packet_len = dp_packet_size(&b),
                     .flow_metadata = MATCH_CATCHALL_INITIALIZER,
