@@ -78,6 +78,8 @@ parse_options(int argc, char *argv[])
         OPT_CLEAR = UCHAR_MAX + 1,
         OPT_MAY_CREATE,
         OPT_READ_ONLY,
+        OPT_NAMES,
+        OPT_NO_NAMES,
         VLOG_OPTION_ENUMS
     };
     static const struct option long_options[] = {
@@ -86,6 +88,8 @@ parse_options(int argc, char *argv[])
         {"may-create", no_argument, NULL, OPT_MAY_CREATE},
         {"read-only", no_argument, NULL, OPT_READ_ONLY},
         {"more", no_argument, NULL, 'm'},
+        {"names", no_argument, NULL, OPT_NAMES},
+        {"no-names", no_argument, NULL, OPT_NO_NAMES},
         {"timeout", required_argument, NULL, 't'},
         {"help", no_argument, NULL, 'h'},
         {"option", no_argument, NULL, 'o'},
@@ -95,6 +99,7 @@ parse_options(int argc, char *argv[])
     };
     char *short_options = ovs_cmdl_long_options_to_short_options(long_options);
 
+    bool set_names = false;
     for (;;) {
         unsigned long int timeout;
         int c;
@@ -125,6 +130,16 @@ parse_options(int argc, char *argv[])
             dpctl_p.verbosity++;
             break;
 
+        case OPT_NAMES:
+            dpctl_p.names = true;
+            set_names = true;
+            break;
+
+        case OPT_NO_NAMES:
+            dpctl_p.names = false;
+            set_names = true;
+            break;
+
         case 't':
             timeout = strtoul(optarg, NULL, 10);
             if (timeout <= 0) {
@@ -156,6 +171,10 @@ parse_options(int argc, char *argv[])
         }
     }
     free(short_options);
+
+    if (!set_names) {
+        dpctl_p.names = dpctl_p.verbosity > 0;
+    }
 }
 
 static void
@@ -177,8 +196,13 @@ usage(void *userdata OVS_UNUSED)
            "  get-flow [DP] ufid:UFID    fetch flow corresponding to UFID\n"
            "  del-flow [DP] FLOW         delete FLOW from DP\n"
            "  del-flows [DP]             delete all flows from DP\n"
-           "  dump-conntrack [DP] [zone=ZONE]  display conntrack entries for ZONE\n"
-           "  flush-conntrack [DP] [zone=ZONE] delete all conntrack entries in ZONE\n"
+           "  dump-conntrack [DP] [zone=ZONE]  " \
+               "display conntrack entries for ZONE\n"
+           "  flush-conntrack [DP] [zone=ZONE] " \
+               "delete all conntrack entries in ZONE\n"
+           "  ct-stats-show [DP] [zone=ZONE] [verbose] " \
+               "CT connections grouped by protocol\n"
+           "  ct-bkts [DP] [gt=N] display connections per CT bucket\n"
            "Each IFACE on add-dp, add-if, and set-if may be followed by\n"
            "comma-separated options.  See ovs-dpctl(8) for syntax, or the\n"
            "Interface table in ovs-vswitchd.conf.db(5) for an options list.\n"
@@ -190,6 +214,7 @@ usage(void *userdata OVS_UNUSED)
            "  -s,  --statistics           print statistics for port or flow\n"
            "\nOptions for dump-flows:\n"
            "  -m, --more                  increase verbosity of output\n"
+           "  --names                     use port names in output\n"
            "\nOptions for mod-flow:\n"
            "  --may-create                create flow if it doesn't exist\n"
            "  --read-only                 do not run read/write commands\n"
